﻿// RMR ("aRMouR") Code Generator
// Copyright 2013 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of RMR Code Generator.

// RMR Code Generator is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// RMR Code Generator is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with RMR Code Generator (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace BenLincoln.RMRCodes
{
    public class MetadataStrip : SVGElement
    {
        public enum MetadataStripDirection
        {
            LSBtoMSB,
            MSBtoLSB
        }

        protected string _StripID;
        protected const int _BitSize = 5;
        protected int _Dimension;
        protected int _NumberOfBits;
        protected byte _Flags;
        protected int _CodeType;
        protected bool _UseParityHorizontal;
        protected bool _UseParityVertical;
        protected bool _IncludeUUID;
        protected MetadataStripDirection _Direction;
        protected MetadataStripBit[] _Bits;

        protected int _BitStepX;
        protected int _BitStepY;
        protected bool _IncludeFlags;

        #region Properties

        public string StripID
        {
            get
            {
                return _StripID;
            }
            set
            {
                _StripID = value;
            }
        }

        public int Dimension
        {
            get
            {
                return _Dimension;
            }
            set
            {
                _Dimension = value;
                RecalculateSize();
                _Bits = new MetadataStripBit[value];
            }
        }

        public byte Flags
        {
            get
            {
                return _Flags;
            }
        }

        public int CodeType
        {
            get
            {
                return _CodeType;
            }
            set
            {
                if ((value < 64) && (value > 0))
                {
                    _CodeType = value;
                    UpdateFlags();
                }
                else
                {
                    throw new Exception("Code type must be between 1 and 63 (inclusive)");
                }
            }
        }

        public bool IncludeUUID
        {
            get
            {
                return _IncludeUUID;
            }
            set
            {
                _IncludeUUID = value;
                UpdateFlags();
            }
        }

        public bool UseParityHorizontal
        {
            get
            {
                return _UseParityHorizontal;
            }
            set
            {
                _UseParityHorizontal = value;
                UpdateFlags();
            }
        }

        public bool UseParityVertical
        {
            get
            {
                return _UseParityVertical;
            }
            set
            {
                _UseParityVertical = value;
                UpdateFlags();
            }
        }

        public MetadataStripDirection Direction
        {
            get
            {
                return _Direction;
            }
            set
            {
                _Direction = value;
            }
        }

        #endregion

        public MetadataStrip()
            : base()
        {
        }

        public MetadataStrip(int xPosition, int yPosition)
            : base(xPosition, yPosition)
        {
        }

        protected override void Initialize()
        {
            base.Initialize();
            StripID = "";
            Dimension = 1;
            Direction = MetadataStripDirection.LSBtoMSB;
            _BitStepX = 10;
            _BitStepY = 10;
            _IncludeFlags = false;
        }

        protected virtual void RecalculateSize()
        {
            _NumberOfBits = 32;
            int currentBit = 31;
            while (currentBit > 0)
            {
                int test = (Dimension >> currentBit) & 1;
                if (test == 1)
                {
                    break;
                }
                currentBit--;
                _NumberOfBits--;
            }
        }

        protected void UpdateFlags()
        {
            _Flags = (byte)(CodeType & 0x1F);
            if (UseParityHorizontal)
            {
                _Flags |= 0x20;
            }
            if (UseParityVertical)
            {
                _Flags |= 0x40;
            }
            if (IncludeUUID)
            {
                _Flags |= 0x80;
            }
        }

        protected virtual MetadataStripBit GetBit()
        {
            return new MetadataStripBit();
        }

        public override string GetSVGCode()
        {
            //RecalculateSize();
            //UpdateFlags();

            string result = GetEmbeddedResourceFileAsString(_EmbeddedResourceFile);

            result = UpdateXPositions(result, BasePositionX);
            result = UpdateYPositions(result, BasePositionY);

            int w = _BitStepX * _NumberOfBits;
            if (w == 0)
            {
                w = 10;
            }
            int h = _BitStepY * _NumberOfBits;
            if (h == 0)
            {
                h = 10;
            }

            int startX = 0;
            int startY = 0;
            int endX = (w - _BitStepX);
            int endY = (h - _BitStepY);
            int stepX = _BitStepX;
            int stepY = _BitStepY;

            if (Direction == MetadataStripDirection.MSBtoLSB)
            {
                if (_BitStepX > 0)
                {
                    startX = (w - _BitStepX);
                }
                if (_BitStepY > 0)
                {
                    startY = (h - _BitStepY);
                }
                endX = 0;
                endY = 0;
                stepX *= -1;
                stepY *= -1;
            }

            StringBuilder bitBuilder = new StringBuilder();

            int currentBit = 0;
            
            long actualValue = Dimension;
            if (_IncludeFlags)
            {
                actualValue <<= 8;
                actualValue |= Flags;
            } 
            
            while (currentBit < _NumberOfBits)
            {
                int currentBitValue = (int)((actualValue >> currentBit) & 1);

                if (currentBitValue == 1)
                {
                    int currentX = BasePositionX + startX + (currentBit * stepX);
                    int currentY = BasePositionY + startY + (currentBit * stepY);
                    MetadataStripBit newBit = GetBit();
                    newBit.BasePositionX = currentX;
                    newBit.BasePositionY = currentY;
                    string bitsvg = newBit.GetSVGCode();
                    bitsvg = bitsvg.Replace("%BIT_NUMBER%", currentBit.ToString());
                    //bitsvg = UpdateXPositions(bitsvg, BasePositionX);
                    //bitsvg = UpdateYPositions(bitsvg, BasePositionY);
                    bitBuilder.Append(bitsvg);
                }

                currentBit++;
            }

            result = result.Replace("%MDS_BITS%", bitBuilder.ToString());
            result = result.Replace("%MDS_ID%", StripID);

            //result = UpdateXPositions(result, BasePositionX);
            //result = UpdateYPositions(result, BasePositionY);

            return result;
        }
    }
}
