﻿// RMR ("aRMouR") Code Generator
// Copyright 2013 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of RMR Code Generator.

// RMR Code Generator is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// RMR Code Generator is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with RMR Code Generator (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

namespace BenLincoln.RMRCodes
{
    public enum RMRCodeType
    {
        Type1 = 1,
        Type2 = 2
    }

    public class RMRCode
    {
        protected const int _SentinelToDataSpacing = 5;
        protected const int _ByteSpacing = 5;
        protected RMRCodeType _CodeType;
        protected int _Width;
        protected int _Height;
        protected int _Length;
        protected int _MinimumHeight;
        protected bool _IncludeUUID;
        protected Guid _UUID;
        protected bool _UseParityHorizontal;
        protected bool _UseParityVertical;
        protected byte[] _Data;

        #region Properties

        public RMRCodeType CodeType
        {
            get
            {
                return _CodeType;
            }
            set
            {
                _CodeType = value;
            }
        }

        public int Width
        {
            get
            {
                return _Width;
            }
            set
            {
                _Width = value;
            }
        }

        public int Height
        {
            get
            {
                return _Height;
            }
        }

        public int Length
        {
            get
            {
                return _Length;
            }
        }

        public int MinimumHeight
        {
            get
            {
                return _MinimumHeight;
            }
            set
            {
                _MinimumHeight = value;
                RecalculateHeight();
            }
        }

        public bool IncludeUUID
        {
            get
            {
                return _IncludeUUID;
            }
            set
            {
                _IncludeUUID = value;
                RecalculateHeight();
            }
        }

        public Guid UUID
        {
            get
            {
                return _UUID;
            }
            set
            {
                _UUID = value;
            }
        }

        public bool UseParityHorizontal
        {
            get
            {
                return _UseParityHorizontal;
            }
            set
            {
                _UseParityHorizontal = value;
                RecalculateHeight();
            }
        }

        public bool UseParityVertical
        {
            get
            {
                return _UseParityVertical;
            }
            set
            {
                _UseParityVertical = value;
                RecalculateHeight();
            }
        }

        public byte[] Data
        {
            get
            {
                return _Data;
            }
            set
            {
                _Data = value;
                RecalculateHeight();
            }
        }

        #endregion

        public RMRCode()
        {
            _CodeType = RMRCodeType.Type1;
            _Width = 3;
            _Height = 1;
            _MinimumHeight = -1;
            _IncludeUUID = false;
            _UUID = new Guid();
            _UseParityHorizontal = false;
            _UseParityVertical = false;
            _Data = new byte[3];
        }

        protected void RecalculateHeight()
        {
            _Length = Data.Length;
            if (IncludeUUID)
            {
                _Length += 32;
            }
            int tempRowCount = (int)(Math.Ceiling((double)((double)_Length / (double)Width)));
            // if horizontal parity is enabled, add one more byte for every row
            if (UseParityHorizontal)
            {
                _Length += tempRowCount;
                tempRowCount = (int)(Math.Ceiling((double)((double)_Length / (double)Width)));
            }
            // if vertical parity is enabled, add one more row
            if (UseParityVertical)
            {
                _Length += Width;
                tempRowCount++;
                //// if horizontal parity is enabled as well, add one more byte
                //if (UseParityHorizontal)
                //{
                //    actualSize++;
                //}
            }
            // if a minimum height has been specified and is greater than the current height, expand to that size
            if (MinimumHeight > tempRowCount)
            {
                _Height = MinimumHeight;
            }
            else
            {
                _Height = tempRowCount;
            }
            _Length = Width * Height;
        }

        public string GetSVGCode()
        {
            RecalculateHeight();
            string result = "";
            RMRCodeFile file = new RMRCodeFile();
            result = file.GetSVGCode();

            int byteWidth;
            int byteHeight;

            ByteType1 b1 = new ByteType1();
            byteWidth = b1.Width;
            byteHeight = b1.Height;

            if (CodeType == RMRCodeType.Type2)
            {
                ByteType2 b2 = new ByteType2();
                byteWidth = b2.Width;
                byteHeight = b2.Height;
            }

            byte[] effData = GetEffectiveData();

            int currentByte = 0;

            int codeGridBaseX;
            int codeGridBaseY;
            int codeGridMaxX;
            int codeGridMaxY;

            StringBuilder contentBuilder = new StringBuilder();
            // add sentinels
            SentinelUpperLeft sUL = new SentinelUpperLeft(0, 0);
            contentBuilder.Append(sUL.GetSVGCode());
            
            codeGridBaseX = sUL.Width;
            codeGridBaseY = sUL.Height;
            codeGridMaxX = codeGridBaseX + (Width * byteWidth);
            codeGridMaxY = codeGridBaseY + (Height * byteHeight);

            SentinelUpperRight sUR = new SentinelUpperRight(codeGridMaxX, 0);
            contentBuilder.Append(sUR.GetSVGCode());

            SentinelLowerRight sLR = new SentinelLowerRight(codeGridMaxX, codeGridMaxY);
            contentBuilder.Append(sLR.GetSVGCode());

            SentinelLowerLeft sLL = new SentinelLowerLeft(0, codeGridMaxY);
            contentBuilder.Append(sLL.GetSVGCode());

            // add metadata strips
            // LSB to MSB horizontal
            MetadataHorizontal mh = new MetadataHorizontal(40, 0);
            mh.StripID = "MDS_ULH";
            mh.Dimension = Width;
            mh.UseParityHorizontal = UseParityHorizontal;
            mh.UseParityVertical = UseParityVertical;
            mh.IncludeUUID = IncludeUUID;
            mh.CodeType = (int)CodeType;
            contentBuilder.Append(mh.GetSVGCode());

            mh.BasePositionY = codeGridMaxY + 30;
            mh.StripID = "MDS_LLH";
            contentBuilder.Append(mh.GetSVGCode());
            
            // MSB to LSB horizontal
            mh.Direction = MetadataStrip.MetadataStripDirection.MSBtoLSB;
            mh.BasePositionX = codeGridMaxX - mh.Width;
            mh.BasePositionY = 15;
            mh.StripID = "MDS_URH";
            contentBuilder.Append(mh.GetSVGCode());

            mh.BasePositionY = codeGridMaxY + 15;
            mh.StripID = "MDS_LRH";
            contentBuilder.Append(mh.GetSVGCode());


            // LSB to MSB vertical
            MetadataVertical mv = new MetadataVertical(0, 40);
            mv.StripID = "MDS_ULV";
            mv.Dimension = Height;
            mv.UseParityHorizontal = UseParityHorizontal;
            mv.UseParityVertical = UseParityVertical;
            mv.IncludeUUID = IncludeUUID;
            mv.CodeType = (int)CodeType;
            contentBuilder.Append(mv.GetSVGCode());

            mv.BasePositionX = codeGridMaxX + 30;
            mv.StripID = "MDS_URV";
            contentBuilder.Append(mv.GetSVGCode());


            // MSB to LSB vertical
            mv.Direction = MetadataStrip.MetadataStripDirection.MSBtoLSB;
            mv.BasePositionY = codeGridMaxY - mv.Height;
            mv.BasePositionX = 15;
            mv.StripID = "MDS_LLV";
            contentBuilder.Append(mv.GetSVGCode());

            mv.BasePositionX = codeGridMaxX + 15;
            mv.StripID = "MDS_LRV";
            contentBuilder.Append(mv.GetSVGCode());


            // add bytes
            for (int y = 0; y < Height; y++)
            {
                int yPos = codeGridBaseY + (y * byteHeight);
                for (int x = 0; x < Width; x++)
                {
                    int xPos = codeGridBaseX + (x * byteWidth);
                    RMRByte b = new RMRByte();
                    switch (CodeType)
                    {
                        case RMRCodeType.Type1:
                            b = new ByteType1(xPos, yPos);
                            break;
                        case RMRCodeType.Type2:
                            b = new ByteType2(xPos, yPos);
                            break;
                        default:
                            throw new Exception("Unsupported code type");
                            break;
                    }
                    b.ByteNumber = currentByte;
                    b.Value = effData[currentByte];
                    currentByte++;
                    contentBuilder.Append(b.GetSVGCode());
                }
            }

            result = result.Replace("%FILE_CONTENTS%", contentBuilder.ToString());
            result = result.Replace("%DOCHEIGHT%", "1000");
            result = result.Replace("%DOCWIDTH%", "1000");
            result = result.Replace("%RMR_CODE_ID%", "101");

            return result;
        }

        protected byte[] GetEffectiveData()
        {
            byte[] result = new byte[Length];

            int currentSourceByte = 0;
            int currentDestByte = 0;

            byte[] columnParities = new byte[Width];
            byte currentRowParity = (byte)0;
            byte overallRowParity = (byte)0;

            for (int row = 0; row < Height; row++)
            {
                currentRowParity = (byte)0;
                if ((!UseParityVertical) || (row < (Height - 1)))
                {
                    for (int col = 0; col < Width; col++)
                    {
                        if ((!UseParityHorizontal) || (col < (Width - 1)))
                        {
                            if (currentSourceByte < (Data.Length - 0))
                            {
                                result[currentDestByte] = Data[currentSourceByte];
                                currentSourceByte++;
                            }
                            else
                            {
                                result[currentDestByte] = (byte)0;
                            }
                            // row parity
                            if (col == 0)
                            {
                                currentRowParity = result[currentDestByte];
                            }
                            else
                            {
                                currentRowParity = (byte)(~(currentRowParity ^ result[currentDestByte]));
                            }
                        }
                        else
                        {
                            //result[currentDestByte] = (byte)0;
                            result[currentDestByte] = currentRowParity;
                        }
                        // column parity
                        if (row == 0)
                        {
                            columnParities[col] = result[currentDestByte];
                        }
                        else
                        {
                            columnParities[col] = (byte)(~(columnParities[col] ^ result[currentDestByte]));
                        }
                        currentDestByte++;
                    }
                    if (row == 0)
                    {
                        overallRowParity = currentRowParity;
                    }
                    else
                    {
                        overallRowParity = (byte)(~(overallRowParity ^ currentRowParity));
                    }
                }
                else
                {
                    for (int col = 0; col < Width; col++)
                    {
                        if ((!(UseParityHorizontal && UseParityVertical)) || (col < (Width - 1)))
                        {
                            result[currentDestByte] = columnParities[col];
                        }
                        else
                        {
                            for (int p = 0; p < (columnParities.Length - 1); p++)
                            {
                                overallRowParity = (byte)(~(overallRowParity ^ columnParities[p]));
                            }
                            result[currentDestByte] = overallRowParity;
                        }
                        currentDestByte++;
                    }
                }
            }

            return result;
        }
    }
}
