﻿// RMR ("aRMouR") Code Generator
// Copyright 2013 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of RMR Code Generator.

// RMR Code Generator is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// RMR Code Generator is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with RMR Code Generator (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

using BenLincoln.RMRCodes;

namespace RMRCodeGenerator
{
    class Program
    {


        static void Main(string[] args)
        {
            int resultCode = 0;

            if ((args.Length < 5))
            {
                DisplayHelpMessage();
                Environment.Exit(1);
            }

            RMRCode rmrCode = new RMRCode();

            string inputFilePath = "";
            string outputFilePath = "";
            bool overwriteExistingOutputFile = false;

            ArrayList arguments = new ArrayList();
            for (int i = 0; i < args.Length; i++)
            {
                // if the argument is one with a second component, catch that now
                switch (args[i])
                {
                    case "-t":
                        try
                        {
                            int cType = int.Parse(args[i + 1]);
                            switch (cType)
                            {
                                case 1:
                                    rmrCode.CodeType = RMRCodeType.Type1;
                                    break;
                                case 2:
                                    rmrCode.CodeType = RMRCodeType.Type2;
                                    break;
                                default:
                                    ExitDueToCommandLine("unsupported RMR code type: " + cType.ToString());
                                    break;
                            }
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToCommandLine("reading the -t (RMR code type) command line value. Please specify either 1 or 2.");
                        }
                        break;
                    case "-U":
                        try
                        {
                            rmrCode.UUID = new Guid((args[i + 1]));
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToCommandLine("reading the -U (explicitly-specified UUID) command line value. Please specify a UUID in any of the formats acceptable to the .NET framework, e.g. \"550e8400-e29b-41d4-a716-446655440000\".");
                        }
                        break;
                    case "-x":
                        try
                        {
                            rmrCode.Width = int.Parse(args[i + 1]);
                            if (rmrCode.Width < 3)
                            {
                                ExitDueToCommandLine("reading the -x (horizontal resolution) command line value. Please specify an integer of at least 3.");
                            }
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToCommandLine("reading the -x (horizontal resolution) command line value. Please specify an integer of at least 3.");
                        }
                        break;
                    case "-y":
                        try
                        {
                            rmrCode.MinimumHeight = int.Parse(args[i + 1]);
                            if (rmrCode.MinimumHeight < 1)
                            {
                                ExitDueToCommandLine("reading the -y (minimum vertical resolution) command line value. Please specify a positive integer.");
                            }
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToCommandLine("reading the -y (minimum vertical resolution) command line value. Please specify a positive integer.");
                        }
                        break;
                    default:
                        arguments.Add(args[i]);
                        break;
                }
            }

            ArrayList remainingArgs = new ArrayList(arguments);

            ArrayList fileList = new ArrayList();

            foreach (string arg in arguments)
            {
                if ((arg == "-?") || (arg == "-h") || (arg == "--help"))
                {
                    DisplayHelpMessage();
                    Environment.Exit(0);
                }
                switch (arg)
                {
                    case "-u":
                        remainingArgs.Remove(arg);
                        rmrCode.IncludeUUID = true;
                        break;
                    case "-ph":
                        remainingArgs.Remove(arg);
                        rmrCode.UseParityHorizontal = true;
                        break;
                    case "-pv":
                        remainingArgs.Remove(arg);
                        rmrCode.UseParityVertical = true;
                        break;
                    case "-f":
                        remainingArgs.Remove(arg);
                        overwriteExistingOutputFile = true;
                        break;
                    default:
                        fileList.Add(arg);
                        remainingArgs.Remove(arg);
                        break;
                }
            }

            if (fileList.Count < 2)
            {
                Console.WriteLine("Error: both an input and an output file must be specified.");
                Environment.Exit(1);
            }

            inputFilePath = (string)fileList[0];
            outputFilePath = (string)fileList[1];

            if (File.Exists(inputFilePath))
            {
                Console.WriteLine("Found input file '" + inputFilePath + "'.");
            }
            else
            {
                ExitWithError("The specified input file ('" + inputFilePath + "') could not be located.", 2);
            }

            if (File.Exists(outputFilePath))
            {
                if (overwriteExistingOutputFile)
                {
                    Console.WriteLine("Existing output file '" + outputFilePath + "' will be overwritten.");
                }
                else
                {
                    ExitWithError("The specified output file ('" + outputFilePath + "') already exists. Use the -f flag to cause it to be overwritten.", 2);
                }
            }
            else
            {
                Console.WriteLine("Using '" + outputFilePath + "' as the output file path.");
            }

            FileInfo f = new FileInfo(inputFilePath);
            int fileLength = (int)f.Length;

            FileStream iStream = new FileStream(inputFilePath, FileMode.Open);
            BinaryReader iReader = new BinaryReader(iStream);
            byte[] data = iReader.ReadBytes(fileLength);
            iReader.Close();
            iStream.Close();
            rmrCode.Data = data;

            FileStream oStream = new FileStream(outputFilePath, FileMode.Create);
            StreamWriter oWriter = new StreamWriter(oStream);

            oWriter.Write(rmrCode.GetSVGCode());

            oWriter.Close();
            oStream.Close();

        }

        static void DisplayHelpMessage()
        {
            Console.WriteLine("RMR (\"Armour\") Code Generator v1.0");
            Console.WriteLine("Copyright 2013 Ben Lincoln");
            Console.WriteLine("http://www.beneaththewaves.net");
            Console.WriteLine(Environment.NewLine);
            Console.WriteLine(Environment.NewLine);
            Console.WriteLine("Usage:");
            Console.WriteLine("\tRMRCodeGenerator.exe -t [1|2] (-u (-U [UUID])) -x [positive integer greater than 2] -y [positive integer] (-ph) (-pv) inputfile outputfile");
            Console.WriteLine(Environment.NewLine);
            Console.WriteLine("-t specifies the code type (currently either 1 or 2)");
            Console.WriteLine(Environment.NewLine);
            Console.WriteLine("-u specifies that the output file contents should have a UUID appended to the beginning and end (this will increase the number of bytes represented by the RMR code by 32)");
            Console.WriteLine("If -u is specified alone, the UUID is randomly-generated. Specify -U and a UUID to use an existing UUID - for example, -U \"550e8400-e29b-41d4-a716-446655440000\".");
            Console.WriteLine(Environment.NewLine);
            Console.WriteLine("-x specifies the x resolution (number of vertical columns) in the output. This must be a positive integer of 3 or greater.");
            Console.WriteLine(Environment.NewLine);
            Console.WriteLine("-y specifies the minimum y resolution (number of horizontal rows) in the output. This must be a positive integer. If this value is lower than the number calculated based on the input data, it will be ignored.");
            Console.WriteLine(Environment.NewLine);
            Console.WriteLine("-ph indicates that the last byte of each row should store the parity of the other bytes in the row. This will increase the output by one byte for every row of data.");
            Console.WriteLine(Environment.NewLine);
            Console.WriteLine("-pv indicates that the last byte of each column should store the parity of the other bytes in the column. This will increase the output by one byte for every column of data.");
            Console.WriteLine(Environment.NewLine);
            Console.WriteLine("Note: if both -ph and -pv are specified, the final byte will store the parity of the horizontal and vertical parity values. This will increase the output by one more byte.");
            Console.WriteLine(Environment.NewLine);
            Console.WriteLine("-f indicates that if the specified output file already exists, it should be overwritten. If the file exists and -f is not specified, no work will be performed.");
            Console.WriteLine(Environment.NewLine);
            Console.WriteLine("The input file can be any file of reasonable size. Null bytes will be appended at the end to pad it as necessary to fill any necessary non-parity bytes.");
            Console.WriteLine(Environment.NewLine);
            Console.WriteLine("The output file will be an SVG image. Use Inkscape or similar tools to convert it to a bitmap if necessary.");
        }

        protected static void ExitDueToCommandLine(string info)
        {
            ExitWithError("An error occurred while " + info + ".", 1);
        }

        protected static void ExitWithError(string message, int exitCode)
        {
            Console.WriteLine(message);
            System.Environment.ExitCode = exitCode;
            Environment.Exit(exitCode);
        }
    }
}
